import warnings
from typing import Any, Dict, Iterable, Optional, Tuple

try:
    # Python 3.8+
    # pylint: disable=ungrouped-imports
    from typing import TypedDict  # type: ignore
except ImportError:
    # Python < 3.8
    from typing_extensions import TypedDict  # type: ignore


class PluginProjectMetaData(TypedDict):
    dist_version: str
    documentation_url: Optional[str]
    download_url: Optional[str]
    source_url: Optional[str]
    tracker_url: Optional[str]


def load_webviz_plugins_with_metadata(
    distributions: Iterable, loaded_plugins: Dict[str, Any]
) -> Tuple[Dict[str, dict], Dict[str, PluginProjectMetaData]]:
    """Loads the given distributions, finds entry points corresponding to webviz-config
    plugins, and put them into the mutable input dictionary loaded_plugins
    (key is plugin name string, value is reference to plugin class).
    Also returns a dictionary of plugin metadata.
    """

    plugin_project_metadata: Dict[str, PluginProjectMetaData] = {}
    plugin_metadata: Dict[str, dict] = {}

    for dist in distributions:
        for entry_point in dist.entry_points:
            if entry_point.group == "webviz_config_plugins":

                dist_name = dist.metadata["name"]

                project_urls = {
                    value.split(",")[0]: value.split(",")[1].strip()
                    for (key, value) in dist.metadata.items()
                    if key == "Project-URL"
                }

                if (
                    entry_point.name in plugin_metadata
                    and dist_name not in plugin_project_metadata
                ):
                    warnings.warn(
                        f"Multiple versions of plugin with name {entry_point.name}. Already "
                        f"loaded from project {plugin_metadata[entry_point.name]['dist_name']}. "
                        f"Overwriting using plugin with from project {dist_name}",
                        RuntimeWarning,
                    )

                plugin_project_metadata[dist_name] = PluginProjectMetaData(
                    {
                        "dist_version": dist.version,
                        "documentation_url": project_urls.get("Documentation"),
                        "download_url": project_urls.get("Download"),
                        "source_url": project_urls.get("Source"),
                        "tracker_url": project_urls.get("Tracker"),
                    }
                )

                plugin_metadata[entry_point.name] = {
                    "dist_name": dist.metadata["name"],
                }

                loaded_plugins[entry_point.name] = entry_point.load()

    return (plugin_metadata, plugin_project_metadata)



import logging
import os
import time
import secrets
import pathlib
import warnings
import webbrowser
import contextlib
import subprocess
from typing import List, Dict, Optional, Tuple, Any

try:
    import azure.cli.core
    from azure.common import AzureHttpError
    from azure.core.exceptions import HttpResponseError
    from msrestazure.azure_exceptions import CloudError

    AZURE_CLI_INSTALLED = True
except ModuleNotFoundError:
    AZURE_CLI_INSTALLED = False

PIMCOMMON_URL = (
    "https://portal.azure.com/#blade/Microsoft_Azure_PIMCommon/ActivationMenuBlade"
)


def _azure_cli(args: List[str], devnull_stderr: bool = True) -> Any:
    """Runs an Azure CLI command using the Azure Python library."""

    # When returning errors, azure cli seems to close the stderr stream similar to
    # https://github.com/pytest-dev/pytest/issues/5502
    # This makes it impossible to catch an error and later perform a new call
    # to azure cli.
    # Therefore cleaning the root logger as suggested in
    # https://github.com/pytest-dev/pytest/issues/5502#issuecomment-647157873
    loggers = [logging.getLogger()] + list(
        logging.Logger.manager.loggerDict.values()  # type: ignore[attr-defined]
    )
    for logger in loggers:
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            logger.removeHandler(handler)

    cli = azure.cli.core.get_default_cli()
    with open(os.devnull, "w") as out_file:
        # contextlib.nullcontext available only in Python 3.7+
        if devnull_stderr:
            with contextlib.redirect_stderr(out_file):
                cli.invoke(args, out_file=out_file)
        else:
            cli.invoke(args, out_file=out_file)

    if cli.result.error:
        raise cli.result.error

    return cli.result.result


def logged_in() -> bool:
    """Returns true if user is logged into Azure,
    otherwise returns False.
    """
    # No easy way currently checking if logged in.
    # https://github.com/Azure/azure-cli/issues/6802

    if not AZURE_CLI_INSTALLED:
        raise RuntimeError(
            "In order to use webviz deploy features, you need to first install "
            "the optional deploy dependencies. You can do this by e.g. running "
            "'pip install webviz-config[deploy]'"
        )

    try:
        _azure_cli(["account", "list-locations"], devnull_stderr=False)
        return True
    except azure.cli.core.CLIError:
        return False


def log_in() -> None:
    _azure_cli(["login", "--use-device-code"])


def subscriptions() -> List[str]:
    """Returns list of all Azure subscriptions logged in user has read access to."""
    return [group["name"] for group in _azure_cli(["account", "list"])]


def resource_groups(subscription: str) -> List[str]:
    """Returns list of all Azure resource group names logged in user has read access to
    within given subscription."""
    try:
        return [
            group["name"]
            for group in _azure_cli(["group", "list", "--subscription", subscription])
        ]
    except CloudError:  # AuthorizationFailed
        return []


def storage_account_name_available(name: str) -> Tuple[bool, str]:
    result = _azure_cli(["storage", "account", "check-name", "--name", name])
    return (result["nameAvailable"], result["message"])


def storage_account_exists(name: str, subscription: str, resource_group: str) -> bool:
    accounts = _azure_cli(
        ["storage", "account", "list", "--subscription", subscription]
    )

    for account in accounts:
        if account["name"] == name:
            if account["resourceGroup"] != resource_group:
                warnings.warn(
                    f"Storage account with name {name} found, but it belongs "
                    f"to another resource group ({account['resourceGroup']}."
                )
            return True

    return False


def storage_container_exists(
    container_name: str, account_name: str, subscription: str
) -> bool:
    containers = _azure_cli(
        [
            "storage",
            "container",
            "list",
            "--account-name",
            account_name,
            "--auth-mode",
            "login",
            "--subscription",
            subscription,
        ]
    )
    return any(container["name"] == container_name for container in containers)


def create_storage_account(subscription: str, resource_group: str, name: str) -> None:
    """Creates an Azure storage account. Also adds upload access, as well
    as possibility to list/generate access keys, to the user creating it
    (i.e. the currently logged in user).

    Note that Azure documentation states that it can take up to five minutes
    after the command has finished until the added access is enabled in practice.
    """

    azure_pim_already_open = False

    while True:
        try:
            _azure_cli(
                [
                    "storage",
                    "account",
                    "create",
                    "--subscription",
                    subscription,
                    "--resource-group",
                    resource_group,
                    "--name",
                    name,
                    "--location",
                    "northeurope",
                    "--sku",
                    "Standard_ZRS",
                    "--encryption-services",
                    "blob",
                ]
            )
            break
        except (HttpResponseError, CloudError) as exc:
            if "AuthorizationFailed" in str(exc):
                if not azure_pim_already_open:
                    webbrowser.open(f"{PIMCOMMON_URL}/azurerbac")
                    print(
                        "Not able to create new storage account. Do you have "
                        "enough priviliges to do it? We automatically opened the URL "
                        "to where you activate Azure PIM. Please activate necessary "
                        "priviliges. You need to be 'Owner' of the subscription in "
                        "order to both create the new account and assign user "
                        "roles to it afterwards."
                    )
                    azure_pim_already_open = True
                print("New attempt of app registration in 1 minute.")
                time.sleep(60)
            else:
                raise RuntimeError("Not able to create new storage account.") from exc

    user_id: str = _azure_cli(
        ["ad", "signed-in-user", "show", "--query", "objectId", "-o", "tsv"]
    )
    resource_group_id: str = _azure_cli(
        ["group", "show", "--subscription", subscription, "--name", resource_group]
    )["id"]

    for role in [
        "Storage Blob Data Contributor",
        "Storage Account Key Operator Service Role",
    ]:
        _azure_cli(
            [
                "role",
                "assignment",
                "create",
                "--role",
                role,
                "--assignee",
                user_id,
                "--scope",
                f"{resource_group_id}/providers/Microsoft.Storage/storageAccounts/{name}",
            ]
        )


def get_storage_account_access_key(subscription: str, account_name: str) -> str:
    result = _azure_cli(
        [
            "storage",
            "account",
            "keys",
            "list",
            "--account-name",
            account_name,
            "--subscription",
            subscription,
        ]
    )

    return result[0]["value"]


def create_storage_container(
    subscription: str, storage_account: str, container: str
) -> None:
    """Creates a container in given storage account."""
    _azure_cli(
        [
            "storage",
            "container",
            "create",
            "--account-name",
            storage_account,
            "--name",
            container,
            "--auth-mode",
            "login",
            "--subscription",
            subscription,
        ]
    )


def storage_container_upload_folder(
    storage_name: str,
    container_name: str,
    destination_folder: str,
    source_folder: pathlib.Path,
) -> None:
    # If the upload access was recently added, Azure documentation
    # says it can take up until five minutes before the access is
    # enabled in practice.

    command_arguments = [
        "storage",
        "blob",
        "upload-batch",
        "--account-name",
        storage_name,
        "--destination",
        f"{container_name}/{destination_folder}",
        "--auth-mode",
        "login",
        "--source",
        str(source_folder),
    ]

    for _ in range(5):
        try:
            _azure_cli(
                command_arguments,
                devnull_stderr=False,  # progress status shown in stderr by Azure CLI...
            )
            return
        except AzureHttpError:
            pass
        except ValueError:
            print(
                "====================================================================\n"
                "Ignore the 'ValueError: I/O operation on closed file' above.\n"
                "There is currently a bug in the Azure Python CLI that hits when you\n"
                " 1) Create a new Azure storage account.\n"
                " 2) Upload to the new storage account in the same Python session.\n\n"
                "We will try to upload in a subprocess.\n"
                "===================================================================="
            )

            result = subprocess.run(  # pylint: disable=subprocess-run-check
                ["az"] + command_arguments,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                # capture_output=True, <-- First vailable in Python 3.7
            )

            if not b"ERROR" in result.stderr:
                # Azure CLI shows progress in stderr,
                # i.e. even successfull runs have content in stderr
                return
        finally:
            print("Waiting on Azure access activation... please wait.")
            time.sleep(60)

    raise RuntimeError("Not able to upload folder to blob storage container.")


def appservice_plan_list() -> List[str]:
    """Returns list of all Azure resource group names logged in user has access to."""
    return [group["name"] for group in _azure_cli(["appservice", "plan", "list"])]


def existing_app_registration(display_name: str) -> Optional[str]:
    existing = _azure_cli(["ad", "app", "list", "--display-name", display_name])
    if existing:
        return existing[0]["appId"]
    return None


def create_app_registration(display_name: str) -> str:

    existing = existing_app_registration(display_name)

    return (
        existing
        if existing is not None
        else _azure_cli(
            [
                "ad",
                "app",
                "create",
                "--display-name",
                display_name,
            ]
        )["appId"]
    )


def create_secret(
    app_registration_id: str, secret_description: str, years: int = 100
) -> str:
    new_secret = secrets.token_urlsafe()
    _azure_cli(
        [
            "ad",
            "app",
            "credential",
            "reset",
            "--append",
            "--id",
            app_registration_id,
            "--credential-description",
            secret_description,
            "--password",
            new_secret,
            "--years",
            str(years),
        ]
    )
    return new_secret


def add_reply_url(app_registration_id: str, reply_url: str) -> None:
    """Will add web reply url to given app registration id, if it does not alredy exist."""

    app_data = _azure_cli(["ad", "app", "show", "--id", app_registration_id])

    reply_urls = app_data["replyUrls"]
    if reply_url not in reply_urls:
        reply_urls.append(reply_url)

    _azure_cli(
        [
            "ad",
            "app",
            "update",
            "--id",
            app_registration_id,
            "--reply-urls",
        ]
        + reply_urls
    )


def create_service_principal(app_registration_id: str) -> Tuple[str, str]:
    existing_service_principal = _azure_cli(
        ["ad", "sp", "list", "--filter", f"appId eq '{app_registration_id}'"]
    )

    service_principal = (
        existing_service_principal[0]
        if existing_service_principal
        else _azure_cli(["ad", "sp", "create", "--id", app_registration_id])
    )

    object_id = service_principal["objectId"]

    _azure_cli(
        [
            "ad",
            "sp",
            "update",
            "--id",
            object_id,
            "--set",
            "appRoleAssignmentRequired=true",
        ]
    )

    return object_id, service_principal["appOwnerTenantId"]


def azure_app_registration_setup(
    display_name: str, proxy_redirect_url: str
) -> Dict[str, str]:

    azure_pim_already_open = False

    while True:
        try:
            app_registration_id = create_app_registration(display_name)
            object_id, tenant_id = create_service_principal(app_registration_id)
            break
        except azure.cli.core.CLIError as exc:

            if not azure_pim_already_open:
                webbrowser.open(f"{PIMCOMMON_URL}/aadmigratedroles")
                azure_pim_already_open = True

                print(exc)
                print(
                    "Not able to create new app registration. Do you have enough priviliges "
                    "to do it? We automatically opened the URL to where you activate Azure PIM. "
                    "Please active necessary priviliges."
                )

            print("New attempt of app registration in 1 minute.")
            time.sleep(60)

    proxy_client_secret = create_secret(app_registration_id, "cli secret")
    add_reply_url(app_registration_id, proxy_redirect_url)

    return {
        "app_registration_id": app_registration_id,
        "object_id": object_id,
        "proxy_client_secret": proxy_client_secret,
        "proxy_cookie_secret": secrets.token_urlsafe(nbytes=16),
        "proxy_redirect_url": proxy_redirect_url,
        "tenant_id": tenant_id,
    }



import glob
import pathlib

from plotly.io import templates

from webviz_config import WebvizConfigTheme

default_theme = WebvizConfigTheme(theme_name="default")  # pylint: disable=invalid-name

default_theme.assets = glob.glob(
    str(pathlib.Path(__file__).resolve().parent / "default_assets" / "*")
)

default_theme.plotly_theme = templates["plotly"].to_plotly_json()



import os
import datetime
from typing import Tuple

import msal
import flask


class Oauth2:
    """Oauth2 authorization Code grant flow"""

    def __init__(self, app: flask.app.Flask):
        self._app = app

        # Azure AD app registration info (currently the values are taken from environment variables)
        self._tenant_id = os.environ["WEBVIZ_TENANT_ID"]
        self._client_id = os.environ["WEBVIZ_CLIENT_ID"]
        self._client_secret = os.environ["WEBVIZ_CLIENT_SECRET"]
        self._scope = os.environ["WEBVIZ_SCOPE"]

        # Initiate msal
        self._msal_app = msal.ConfidentialClientApplication(
            client_id=self._client_id,
            client_credential=self._client_secret,
            authority=f"https://login.microsoftonline.com/{self._tenant_id}",
        )
        self._accounts = self._msal_app.get_accounts()

        # Initiate oauth2 endpoints
        self.set_oauth2_endpoints()

    def set_oauth2_endpoints(self) -> None:
        """/login and /auth-return endpoints are added for Oauth2 authorization
        code flow.

        At the end of the flow, a session cookie containing a valid access token
        and its expiration date is created. This flask session object can be
        accessed from Webviz plugin.

        To get the access token: flask.session.get("access_token")
        To get the expiration date: flask.session.get("expiration_date")

        An Azure AD application should be registered, and the following environment
        variables should be set: WEBVIZ_TENANT_ID, WEBVIZ_CLIENT_ID,
        WEBVIZ_CLIENT_SECRET, WEBVIZ_SCOPE.
        """

        @self._app.route("/login")
        def _login_controller():  # type: ignore[no-untyped-def]
            redirect_uri = get_auth_redirect_uri(flask.request.url_root)

            # First leg of Oauth2 authorization code flow
            auth_url = self._msal_app.get_authorization_request_url(
                scopes=[self._scope], redirect_uri=redirect_uri
            )
            return flask.redirect(auth_url)

        @self._app.route("/auth-return")
        def _auth_return_controller():  # type: ignore[no-untyped-def]
            redirect_uri = get_auth_redirect_uri(flask.request.url_root)
            returned_query_params = flask.request.args

            # There is an error from the first leg of Oauth2 authorization code flow
            if "error" in returned_query_params:
                error_description = returned_query_params.get("error_description")
                print("Error description:", error_description)
                redirect_error_uri = flask.url_for("error_controller")
                return flask.redirect(redirect_error_uri)

            code = returned_query_params.get("code")

            # Second leg of Oauth2 authorization code flow
            tokens_result = self._msal_app.acquire_token_by_authorization_code(
                code=code, scopes=[self._scope], redirect_uri=redirect_uri
            )
            expires_in = tokens_result.get("expires_in")
            expiration_date = datetime.datetime.now() + datetime.timedelta(
                seconds=expires_in - 60
            )
            print("Access token expiration date:", expiration_date)

            # Set expiration date in the session
            flask.session["expiration_date"] = expiration_date

            # Set access token in the session
            flask.session["access_token"] = tokens_result.get("access_token")

            return flask.redirect(flask.request.url_root)

        @self._app.route("/error")
        def _error_controller():  # type: ignore[no-untyped-def]
            return "Error"

    def set_oauth2_before_request_decorator(self) -> None:
        """Check access token existence in session cookie before every request.
        If it does not exist, the browser is redirected to /login endpoint.

        If access token exists, its expiration date is checked in session cookie.
        If the current date exceeds its expiration date, a new access token is
        retrieved and set in the session cookie. A new expiration date is also
        set in the session cookie.
        """

        # pylint: disable=inconsistent-return-statements
        @self._app.before_request
        def _check_access_token():  # type: ignore[no-untyped-def]
            # The session of the request does not contain access token, redirect to /login
            is_redirected, redirect_url = self.is_empty_token()
            if is_redirected:
                return flask.redirect(redirect_url)

            # The session contains access token, check its expiration date
            self.check_and_set_token_expiry()

    @staticmethod
    def is_empty_token() -> Tuple[bool, str]:
        if (
            not flask.session.get("access_token")
            and flask.request.path != "/login"
            and flask.request.path != "/auth-return"
        ):
            login_uri = get_login_uri(flask.request.url_root)
            return True, login_uri

        return False, ""

    def check_and_set_token_expiry(self) -> None:
        if flask.session.get("access_token"):
            expiration_date = flask.session.get("expiration_date")
            current_date = datetime.datetime.now()
            if current_date > expiration_date:
                # Access token has expired
                print("Access token has expired.")
                if not self._accounts:
                    self._accounts = self._msal_app.get_accounts()
                renewed_tokens_result = self._msal_app.acquire_token_silent(
                    scopes=[self._scope], account=self._accounts[0]
                )
                expires_in = renewed_tokens_result.get("expires_in")
                new_expiration_date = datetime.datetime.now() + datetime.timedelta(
                    seconds=expires_in - 60
                )
                print("New access token expiration date:", new_expiration_date)

                # Set new expiration date in the session
                flask.session["expiration_date"] = new_expiration_date

                # Set new access token in the session
                flask.session["access_token"] = renewed_tokens_result.get(
                    "access_token"
                )


def get_login_uri(url_root: str) -> str:
    return url_root + "login"


def get_auth_redirect_uri(url_root: str) -> str:
    return url_root + "auth-return"



import sys
import getpass
import datetime
import pathlib
import argparse
from typing import Dict, Tuple

import jinja2

from ._config_parser import ConfigParser


def write_script(
    args: argparse.Namespace,
    build_directory: pathlib.Path,
    template_filename: str,
    output_filename: str,
) -> Tuple[set, Dict[str, dict]]:
    """Writes rendered script to build directory. Also returns information regarding
    assets and which plugins that are incluced in the user provided configuration file.
    """

    config_parser = ConfigParser(args.yaml_file)
    configuration = config_parser.configuration

    configuration["shared_settings"] = config_parser.shared_settings
    configuration["portable"] = args.portable is not None
    configuration["loglevel"] = args.loglevel
    configuration["config_folder"] = repr(args.yaml_file.resolve().parent)

    configuration["theme_name"] = args.theme

    configuration["author"] = getpass.getuser()
    configuration["current_date"] = datetime.date.today().strftime("%Y-%m-%d")
    configuration["sys_executable"] = sys.executable

    template_environment = jinja2.Environment(  # nosec
        loader=jinja2.PackageLoader("webviz_config", "templates"),
        undefined=jinja2.StrictUndefined,
        autoescape=False,
    )

    template = template_environment.get_template(template_filename)

    (build_directory / output_filename).write_text(template.render(configuration))

    return config_parser.assets, config_parser.plugin_metadata



import pathlib

import yaml
import jsonschema

from webviz_config._docs._create_schema import create_schema


def test_schema():
    """Tests both that the generated schema is valid,
    and that the input configuration is valid according to the schema.
    """

    config = yaml.safe_load(
        (pathlib.Path("examples") / "basic_example.yaml").read_text()
    )
    jsonschema.validate(instance=config, schema=create_schema())
